(** Conservative variant of the lisp collector. *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export rangeset.
Require Import machine2.
Require Export seplog.
Require Import reach.

(** 1. the objects are fixed to have 2 cells. *)
(** 2. there is no aux encoding to separate value from pointer *)
(** 3. the mark_field function take a two step pointer check before the mark bit check. *)
(** 4. the notion of reachability is conservative. *)

(** heap layout: *)

(** object heap                        mark bits *)
(**   |----|----|----|----|......        |----|----|     *)
(** ost    +8                          oed    +4 *)

(** ost, oed align at 8.  *)

(** valid object pointer set:  *)
(**   { x | ost <= x < oed /\ x mod 8 = 0 } *)

(** object pointers align at 8. *)

(** so, mark bit for pointer x:  *)
(**   (div2 (x - ost)) + oed. *)



(** ******************************************************************* *)
(** Constant values *)
(** ******************************************************************* *)
Notation NULL := 0.
Notation HDSIZE := 4.
Notation UNMARK := 0.
Notation MARKED := 1.


(** ******************************************************************* *)
(** labels and registers *)
(** ******************************************************************* *)
Notation rost := t0.
Notation roed := t1.

(** registers *)
Notation rtemp := ast.
Notation rtop := t2.
Notation rbot := t3.
Notation rbuff := t4.

Notation rsweep := a0.
Notation rfree := t5.

(** Control stack *)
Notation rcst1 := t6.
Notation rcst2 := t7.
Notation rcst3 := t8.

Notation rtemp2 := t9.

(** Various Returns *)
Notation RETURN := 0.
Notation RETURN1 := 1.

(** markField() *)
Notation MARKFIELD  := 2.

(** mark() *)
Notation MARK := 3.
Notation MARK1 := 4.
Notation MARK2 := 5.
Notation MARKLOOP := 6.
Notation MARKLOOPBODY := 7.
Notation MARKFIRST := 8.
Notation MARKSECOND := 9.

(** sweep() *)
Notation SWEEP := 10.
Notation SWEEPLOOP := 11.
Notation SWEEPNEXT := 12.
Notation SWEEPADD := 13.
Notation SWEEPFIN := 14.

(** markSweepGC() *)
Notation GC := 15.
Notation GC2 := 16.

(** alloc() *)
Notation ALLOC := 17.
Notation ALLOCLOOP := 18.
Notation ALLOCRETPTR := 19.

(** mark stack operations *)
Notation STACKEMPTY := 20.
Notation STACKEMPTYT := 21.
Notation STACKPUSH := 22.
Notation LOOP := 23.
Notation STACKPOP := 24.


(** ******************************************************************* *)
(** mark stack as set *)
(** ******************************************************************* *)
Notation "s |||- A" := (hp_of s ||- A) (at level 80).

Definition buffer (x y : nat) : hpred :=
  foralls z in addrs x y, z |->_ .

Lemma buffer_remove_head : forall x y h A,
  x + 4 <= y -> h ||- buffer x y ** A -> h ||- x |->_ ** buffer (x + 4) y ** A.
Proof.
  unfold buffer; intros.
  replace (addrs (x + 4) y) with (set_remove x (addrs x y)).
  apply fas_elim with (P := fun x => x |->_); trivial.
  apply addrs_intro with 0; omega.
  apply addrs_remove_min.
Qed.

Lemma buffer_add_head : forall x y n h,
  x >= 4 -> x <= y -> h ||- x - 4 |-> n ** buffer x y -> h ||- buffer (x - 4) y.
Proof.
  unfold buffer; intros.
  replace (addrs (x - 4) y) with (set_add (x - 4) (addrs x y)).
  apply fas_intro1.
  lexintro 1; exists n; trivial.
  intro F.
  destruct (addrs_elim F) as [_ [k Hk]].
  omega.
  apply set_ext; split; intros.
  destruct (set_add_elim H2).
  apply addrs_intro with 0; omega.
  decompose [and ex] (addrs_elim H3).
  apply addrs_intro with (x0 + 1); omega.
  decompose [and ex] (addrs_elim H2).
  destruct x0.
  replace n0 with (x - 4); 
    [apply set_add_intro1 | omega].
  apply set_add_intro2.
  apply addrs_intro with x0; omega.
Qed.

Fixpoint array_set (h : nat) (elems : list nat) {struct elems} : hpred :=
  match elems with
    | nil => unit
    | x::elems' => h |-> x ** array_set (h - WSIZE) elems'
  end.

Definition buffer_array_stack x y z E :=
  !(x + WSIZE * length E = y /\ y <= z) **
  array_set (y - WSIZE) E ** buffer y z.

Definition stack_hp objs s :=
  buffer_array_stack (rget s rbot) (rget s rtop) (rget s rbuff) objs.

Notation stack_pre :=
  (fun s => exists objs, s |||- stack_hp objs s ** top).

Notation sbool := (fun n P => match n with O => ~P | _ => P end).

(** set implemented on top of a stack *)
Definition buffer_stack_set x y z objs :=
  lexists E, !(sd_eqv (sd_of objs) E /\ length (sd_of objs) = length E) **
    buffer_array_stack x y z E.

Definition set_hp objs s :=
  buffer_stack_set (rget s rbot) (rget s rtop) (rget s rbuff) objs.


(** ******************************************************************* *)
(** The set of mark-sweep objects *)
(** ******************************************************************* *)
Lemma ms_fun_ok : forall regMin x y,
  x < y -> (fun k => regMin + 8 * k) x < (fun k => regMin + 8 * k) y.
Proof.
  intros.
  omega.
Qed.

Definition ms_objs (min max: nat) :=
  set_map (set0n (div8 (max - min))) (ms_fun_ok min).

Lemma ms_objs_intro : forall n st ed k,
  ed >= n + 8 -> n = st + 8 * k ->
  set_in n (ms_objs st ed).
Proof.
  intros.
  subst.
  unfold ms_objs.
  eapply set_map_intro.
  reflexivity.
  apply set0n_intro1.
  apply div8_intro_lt.
  omega.
Qed.

Lemma ms_objs_elim : forall n st ed,
  set_in n (ms_objs st ed) ->
  ed >= n + 8 /\ exists k, n = st + 8 * k.
Proof.
  intros a x y I.
  hypreplace I (set_map_elim I).
  destruct I as [b [bEq I]].
  subst.
  constructor.
  hypreplace I (set0n_elim1 I).
  assert (8 * S b <= y - x).
  2:omega.
  assert (X := div8_elim_lt I).
  omega.
  exists b; reflexivity.
Qed.

Lemma ms_objs_subset : forall st ed ed',
  ed <= ed' ->
  set_subset (ms_objs st ed) (ms_objs st ed').
Proof.
  red; intros.
  destruct (ms_objs_elim H0) as [X1 [k Hk]].
  apply ms_objs_intro with k; auto.
  omega.
Qed.


(** ******************************************************************* *)
(** conservative object heap *)
(** ******************************************************************* *)
Notation ohp_ptrs := (ms_objs ost oed).

Notation vptr' := (fun x => set_in x ohp_ptrs).

Lemma vptr_in_ohp_ptrs : forall x,
  vptr x -> set_in x ohp_ptrs.
Proof.
  intros x [R M].
  destruct (mod8_elim M) as [k Hk].
  unfold ost in * ; unfold oed in * .
  apply ms_objs_intro with (k - (S ohpst));
    omega.
Qed.

Lemma in_ohp_ptrs_vptr : forall x,
  set_in x ohp_ptrs -> vptr x.
Proof.
  intros x In.
  destruct (ms_objs_elim In) as [R [k K]].
  split.
  omega.
  unfold ost in K.
  apply mod8_intro with (S ohpst + k).
  omega.
Qed.

(* if x is in the valid pointer set, x will be a pointer in s2 *)
Definition ok_val (s : set) (x : nat) : Prop := set_in x ohp_ptrs -> set_in x s.

Definition ok_fld (s : set) (x:nat) : hpred :=
  lexists y, ! ok_val s y ** x |-> y.

Definition ok_obj (s : set) (x:nat) : hpred :=
  (ok_fld s x) ** (ok_fld s (x + 4)).

Definition obj_hp (s objs:set) : hpred := 
  foralls x in objs, ok_obj s x.

Lemma obj_hp_subset : forall h ss ss' objs,
  h ||- obj_hp ss objs ->
    set_subset ss ss' ->
    h ||- obj_hp ss' objs.
Proof.
  unfold obj_hp; intros.
  apply fas_impl with (P := fun x => ok_obj ss x)
    (P' := fun x => ok_obj ss' x); trivial.
  unfold ok_obj; unfold ok_fld; unfold ok_val; intros.
  sepelim H1.
  lexintro 1; exists x.
  lexintro 2; exists x0.
  simplsep 2 2.
Qed.

Lemma obj_hp_close : forall h x y ss,
  h ||- obj_hp ss ss ** top -> 
  set_in x ss ->
  reach h x y ->
  set_in y ss.
Proof.
  intros h x y ss Hh In Rch.
  induction Rch; trivial.
  generalize (fas_elim In Hh).
  unfold ok_obj; unfold ok_fld; unfold ok_val; intro Hh'; sepelim Hh'.
  assert (Z := vptr_in_ohp_ptrs (reach_vptrx Rch)).
  destruct H as [V | V].
  rewrite (cont_get_eq1 Hh') in V; injection V; intros; subst; auto.
  sconjlifth 1 Hh'.
  rewrite (cont_get_eq1 Hh') in V; injection V; intros; subst; auto.
Qed.

Lemma obj_hp_subheap_get : forall h h' x ss,
  h ||- hp_eq h' ** top -> 
  h' ||- obj_hp ss ss ->
  set_in x ss ->
  heap_get h x = heap_get h' x /\ heap_get h (x + 4) = heap_get h' (x + 4).
Proof.
  intros h h' x ss Hp Hh In.
  destruct Hp as [ha [hb [J [Eq _]]]].
  unfold hpred_true in Eq; unfold hp_eq in Eq; subst h'.
  generalize (fas_elim1 In Hh).
  unfold ok_obj; unfold ok_fld; unfold ok_val; intro Hh'; sepelim Hh'.
  assert (heap_get ha x = Some x0).
  apply (cont_get_eq1 Hh').
  assert (heap_get ha (x + 4) = Some x1).
  sconjlifth 1 Hh'.
  apply (cont_get_eq1 Hh').
  split.
  rewrite (heap_join_sound_left J H); auto.
  rewrite (heap_join_sound_left J H0); auto.
Qed.

Definition some_obj x := x |->_ ** (x + 4) |->_.

Definition some_hp ss := foralls x in ss, some_obj x.

Lemma some_hp_obj_hp : forall h,
  h ||- some_hp ohp_ptrs -> h ||- obj_hp ohp_ptrs ohp_ptrs.
Proof.
  unfold obj_hp; unfold some_hp; intros h Hh.
  eapply fas_impl; eauto.
  unfold ok_obj; unfold ok_fld; 
    unfold ok_val; unfold some_obj; intros.
  sepelim H.
  lexintro 1; exists x.
  lexintro 1; exists x0; simplsep 4 2.
Qed.

Lemma obj_hp_some_hp : forall h ss ptrs,
  h ||- obj_hp ss ptrs -> h ||- some_hp ptrs.
Proof.
  unfold obj_hp; unfold some_hp; unfold some_obj;
    unfold ok_obj; unfold ok_val; unfold ok_fld; intros.
  eapply fas_impl; eauto; intros.
  simpl in H0; sepelim H0.
  lexintro 1; exists x.
  lexintro 1; exists x0.
  simplsep 2 2.
Qed.

Lemma some_hp_set_ok : forall s h l v v',
  h ||- some_hp s ->
    heap_get h l = Some v ->
    heap_set h l v' ||- some_hp s.
Proof.
  destruct s.
  induction s as [| x s].
  intros.
  assert (h ||- unit).
  unfold some_hp in H.
  apply fas_elim2 with (P := (fun x => some_obj x)); auto.
  unfold hpred_true in H1; unfold unit in H1.
  rewrite <- (heap_ext H1) in H0.
  discriminate.
  intros.
  assert (In : set_in x (pfset e)).
  left; reflexivity.
  hypreplace H (fas_elim1 In H). 
  destruct H as [ha [hb [J [Ha Hb]]]].
  unfold some_obj in Ha; sepelim Ha.
  assert (E := e).
  simpl in E; destruct (andb_prop _ _ E) as [Lt e'].
  replace (pfset e) with (set_add x (pfset e')).
  unfold some_hp.
  apply fas_intro1.
  replace (set_remove x (pfset e)) with (pfset e') in Hb.
  compare l x; intro E1; 
    compare l (x + 4); intro E2; subst.
  elimtype False; omega.
  assert (V : heap_get ha x = Some x0).
  apply (cont_get_eq1 Ha).
  rewrite (heap_join_sound_left J V) in H0; injection H0; intros; subst x0; clear H0.
  hypreplace Ha (cont_update1 Ha v').
  hypreplace J (heap_join_set_left v' V J).
  apply sconj_intro with (heap_set ha x v') hb; trivial.
  unfold some_obj.
  lexintro 1; exists v'.
  lexintro 1; exists x1.
  simplsep 2 2.

  assert (V : heap_get ha (x + 4) = Some x1).
  sconjlifth 1 Ha.
  apply (cont_get_eq1 Ha).
  rewrite (heap_join_sound_left J V) in H0; injection H0; intros; subst x1; clear H0.
  sconjlifth 1 Ha.
  hypreplace Ha (cont_update1 Ha v').
  hypreplace J (heap_join_set_left v' V J).
  apply sconj_intro with (heap_set ha (x + 4) v') hb; trivial.
  unfold some_obj.
  lexintro 1; exists x0.
  lexintro 1; exists v'.
  simplsep 2 2.

  assert (heap_get ha l = None).
  destruct Ha as [h1 [h2 [Ja [[P1 V1] [P2 V2]]]]].
  unfold singleton_heap in * .
  apply heap_join_sound_none with h1 h2; subst; auto.
  simplhp; reflexivity.
  simplhp; reflexivity.
  destruct (heap_get_dec2 hb l) as [B | [vv B]].
  rewrite (heap_join_sound_none J H B) in H0; discriminate.
  apply sconj_intro with ha (heap_set hb l v'); trivial.
  unfold some_obj; lexintro 1; exists x0; lexintro 1; exists x1; trivial.
  apply IHs with vv; trivial.
  apply heap_join_sym.
  apply heap_join_set_left with vv; trivial.
  apply heap_join_sym; trivial.

  apply set_ext; split; unfold set_in; simpl; simplbeq; intros.
  apply sd_remove_intro2; trivial.
  intro F; subst.
  apply (lt_sd_not_in (blt_sd_elim Lt) H).
  destruct (sd_remove_elim H); trivial.
  apply (lt_sd_not_in (blt_sd_elim Lt)).

  apply set_ext; split; unfold set_in; simpl; intros.
  destruct (sd_insert_elim H); subst; [left | right]; trivial.
  destruct H; subst.
  apply sd_insert_intro1.
  apply sd_insert_intro2; trivial.
Qed.


(** ******************************************************************* *)
(** mark-sweep headers *)
(** ******************************************************************* *)
Definition obj_hdrs objs n := 
  foralls x in objs, (oed + div2 (x - ost)) |-> n.

Definition bobj_hdrs objs p m n := 
  foralls x in objs, 
  (oed + div2 (x - ost)) |-> (if blt_nat x p then m else n).

Notation hdr_addr := (fun x => oed + div2 (x - ost)).

Lemma obj_hdrs_merge : forall s1 s2 h n,
  h ||- obj_hdrs s1 n ** obj_hdrs s2 n ->
    h ||- obj_hdrs (set_union s1 s2) n.
Proof.
  intros [s e].
  induction s; simpl; intros.
  replace (set_union (pfset (s:=nil) e) s2) with s2.
  apply (unit_elim_left H).
  apply set_ext; split; intros; simplset.
  inversion SS.
  simpl in e.
  destruct (andb_true_eq _ _ (sym_eq e)).
  replace (pfset (s:=a :: s) e) with (set_add a (pfset (sym_eq H1))).
  replace (set_union (set_add a (pfset (s:=s) (sym_eq H1))) s2) with
    (set_add a (set_union (pfset (s:=s) (sym_eq H1)) s2)).
  unfold obj_hdrs.
  apply fas_intro1.
  unfold obj_hdrs in * ; unfold fas in * ; simpl in * .
  simplsep 3 3.
  apply IHs; auto.
  intro F.
  destruct (set_union_elim F).
  hypreplace H0 (lt_sd_not_in (blt_sd_elim (sym_eq H0))).
  tauto.
  unfold obj_hdrs in H.
  hypreplace H (fas_elim H2 (sconj_sym H)).
  unfold obj_hdrs in H; unfold fas in H; simpl in H.
  apply cont_cont_false with h (hdr_addr a) n n top.
  simplsep 3 3.
  setsolver.
  setsolver.
  subst n0.
  left; reflexivity.
  right; apply SS.
  destruct H2; subst; simplset.
Qed.

Lemma bobj_hdrs_elim : forall s h m n p,
  (forall x, set_in x s -> x < p) ->
  h ||- bobj_hdrs s p m n ->
    h ||- obj_hdrs s m.
Proof.
  unfold bobj_hdrs; unfold obj_hdrs; intros.
  eapply fas_impl1; intros.
  apply H0.
  simpl in H2.
  replace (if blt_nat a p then m else n) with m in H2; trivial.
  simplif; trivial.
  hypreplace H (H _ H1).
  omega.
Qed.

Lemma bobj_hdrs_intro : forall s h m n p,
  (forall x, set_in x s -> x >= p) ->
  h ||- obj_hdrs s n ->
  h ||- bobj_hdrs s p m n.
Proof.
  unfold bobj_hdrs; unfold obj_hdrs; intros.
  eapply fas_impl1; eauto; intros.
  simplif; trivial.
  hypreplace H (H _ H1).
  elimtype False; omega.
Qed.


(** ******************************************************************* *)
(** free list *)
(** ******************************************************************* *)
Fixpoint free_list' (x : nat) (objs : set_data) {struct objs} : hpred :=
  match objs with
    | nil => !(x = NULL)
    | a::objs' => lexists x', !(x = a) ** !(x <> NULL) **
	(a |->_ ** (WSIZE + a) |-> x') ** free_list' x' objs'
  end.

Definition free_list (x : nat) (objs : set) :=
  lexists E, !(sd_eqv (sd_of objs) E /\ length (sd_of objs) = length E) **
  free_list' x E.


Lemma free_list_intro : forall s x y h A,
  x <> NULL -> h ||- x |->_ ** 4 + x |-> y ** free_list y s ** A ->
    ~ set_in x s -> h ||- free_list x (set_add x s) ** A.
Proof.
  induction s.
  unfold free_list; simpl.
  intros x y h A Ne; intros; sepelim H.
  destruct A0.

  lexintro 1; exists (x :: x1).
  simpl.
  lexintro 4; exists y.
  lexintro 4; exists x0.
  simplsep 7 5.
  sconjlift 1; simplsep 1 1.
  split.
  split; intros; destruct (H1 n).
  destruct (sd_insert_elim H3).
  subst n; left; auto.
  right; auto.
  destruct H3.
  subst n.
  apply sd_insert_intro1; auto.
  apply sd_insert_intro2; auto.
  
  hypreplace H0 (set_add_length_inc H0).
  simpl in H0; omega.
  sconjlift 1; simplsep 2 2.
Qed.

Lemma free_list_remove : forall s x h A,
  x <> 0 -> 
  h ||- free_list x s ** A ->
    exists v, exists y, 
      h ||- x |-> v ** 4 + x |-> y ** free_list y (set_remove x s) ** A /\
	set_in x s.
Proof.
  destruct s.
  unfold free_list; simpl.
  intros addr h A Neq0 Hp; sepelim Hp.

  destruct A0 as [Heqv Hlen].
  destruct x as [| addr' x] ; unfold free_list' in Hp; 
    fold free_list' in Hp; sepelim Hp; subst; try tauto.
  exists x1; exists x0.
  split.
  lexintro 4; exists x; simplsep 3 3.
  split.
  split; intros.
  destruct (sd_remove_elim H) as [X1 X2].
  destruct (Heqv n) as [X _].
  hypreplace X (X X2).
  destruct X; trivial; tauto.
  compare n addr'; intro E; subst.
  elimtype False.
  generalize H.
  apply sd_eqv_sd_no_dup with (s := (pfset e)); trivial.
  apply sd_remove_intro2; trivial.
  destruct (Heqv n).
  apply H1; right; trivial.
  assert (set_in addr' (pfset e)).
  destruct (Heqv addr').
  apply H0; left; reflexivity.
  hypreplace H (set_remove_length_dec H).
  simpl in * .
  rewrite Hlen in H; inversion_clear H.
  reflexivity.
  destruct (Heqv addr').
  apply H0; left; reflexivity.
Qed.

Lemma free_list_elim0 : forall s h A,
  h ||- free_list NULL s ** A ->
    h ||- A /\ s = empty_set.
Proof.
  intros s h A Hp; subst.
  unfold free_list in Hp; sepelim Hp.
  destruct x; simpl in Hp; sepelim Hp; destruct A0.
  split; trivial.
  apply set_ext.
  apply sd_eqv_elim.
  simpl; trivial.
  destruct A2; reflexivity.
Qed.


(** ******************************************************************* *)
(** aux definitions *)
(** ******************************************************************* *)
Notation tal_regs'' := (s1::s2::rcst3::nil).

Definition roots_pred (P : nat -> Prop) s :=
  forall r, rlst_in r tal_regs'' -> P (rget s r).

Notation roots_ok := (fun (ptrs : set) s => 
  (roots_pred (fun v => ok_val ptrs v) s)).

Notation sted_ok := (fun s => (rget s rost = ost /\ rget s roed = oed)).

Notation reg_nin_eq := (fun rl s s' => 
  (forall r, rlst_nin r rl -> rget s' r = rget s r)).

Lemma roots_pred_next : forall s s' P,
  roots_pred P s ->
  (forall r, rlst_in r tal_regs'' -> rget s r = rget s' r) ->
  roots_pred P s'.
Proof.
  unfold roots_pred; simpl; intros.
  intuition; subst; rewrite <- H0; intuition.
Qed.  

Ltac regin := repeat split; try discriminate; trivial.

Notation flst_hp := (fun ss s => free_list (rget s rfree) ss).


(** ******************************************************************* *)
(** invariant needed for the safe execution of gc *)
(** ******************************************************************* *)
Definition gc_inv s h := exists B, exists F,
  sted_ok s /\ set_union B F = ohp_ptrs /\
  s |||- hp_eq h ** flst_hp F s ** obj_hdrs ohp_ptrs UNMARK ** set_hp empty_set s /\
  h ||- some_hp B.

Notation tal_regs := (s1::s2::zero::ra::nil).

Lemma gc_inv_rf_set_ok : forall s h r v,
  rlst_in r tal_regs ->
  gc_inv s h -> gc_inv (rset s r v) h.
Proof.
  intros s h r v Hr [B [F [[Hst Hed] [Hobjs [Hp Hh]]]]].
  exists B; exists F.
  destruct Hr as [R |[R |[R |[R | R]]]]; subst; try tauto.
Qed.

Lemma gc_inv_heap_set_ok : forall s h v v' l,
  gc_inv s h ->
  heap_get h l = Some v ->
  gc_inv (hset s l v') (heap_set h l v').
Proof.
  intros s h v v' l Inv Hl.
  destruct Inv as [B [F [[Hst Hed] [Hobjs [Hp Hh]]]]].
  exists B; exists F; repeat split; trivial.
  destruct Hp as [h' [ht [J [Eq Ht]]]].
  unfold hp_eq in Eq; unfold hpred_true in Eq; subst h'.
  hypreplace J (heap_join_set_left v' Hl J).
  apply sconj_intro with (heap_set h l v') ht; trivial.
  unfold hp_eq; unfold hpred_true; reflexivity.
  apply some_hp_set_ok with v; trivial.
Qed.