(** Lightweighted OCAP. *)

(** Rules and Soundness. *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export tal.
Require Export cgc.


(** ******************************************************************* *)
(** interpretation function *)
(** ******************************************************************* *)

(** TAL *)

(* relation R: domain of ha, chspec, hb, rf *)
Notation R := (fun ha (Psi : chspec) s => gc_inv s ha).

Definition interp1 G : chstpred :=
  fun P s => 
    let (rf, hp) := s in
      (exists ha, exists hb,
	heap_join ha hb hp /\
	WFst P G (rf, ha) /\ R ha P s).


(** SCAP *)
Fixpoint WFST (n:nat) (q:spred) (P:chspec) {struct n} : Prop :=
  forall s', q s' ->
    match n with
      | 0 => exists G, (codeptr (rget s' r31) (isp1 G) //\\ (interp1 G)) P s'
      | S n' => exists p', exists g', 
	chspec_in P (rget s' r31) (isp2 p' g') /\
	p' s' /\ WFST n' (g' s') P
    end.

Definition interp2 p g : chstpred :=
  fun P => fun s => p s /\ exists n, WFST n (g s) P.

Lemma WFST_weakening : forall n (q q' : spred) P,
  (forall s, q s -> q' s) -> WFST n q' P -> WFST n q P.
Proof.
  induction n; red; intros;
    apply (H0 s' (H s' H1)).
Qed.
  
Definition interp ispec : chstpred :=
  match ispec with
    | isp1 G => interp1 G
    | isp2 p g => interp2 p g
  end.


(** ******************************************************************* *)
(** typing rules *)
(** ******************************************************************* *)

(** well-formed command rule *)
Definition comm_ok (a a' : chstpred) (c : comm) : Prop :=
  a =>> (fun P s => opt_pred (cstep c s) (fun s' => a' P s')).
  
(** well-formed instruction sequence rule *)
Fixpoint iseq_ok (a : chstpred) (is : iseq) {struct is} : Prop :=
  match is with
    | seq c is' => 
      exists a', comm_ok a a' c /\ iseq_ok a' is'
    | br bt rs rt offset is' => 
      exists a', iseq_ok a' is' /\
	a =>> (fun P s => 
          if (btest bt (rget s rs) (rget s rt))
	    then (exists pi, (codeptr offset pi //\\ interp pi) P s) 
	      else a' P s)
    | j target => 
      a =>> (fun P s => exists pi, (codeptr target pi //\\ interp pi) P s) 
    | jal fret target => 
      a =>> (fun P s => exists pi, (codeptr target pi //\\ interp pi) P (rset s r31 fret)) 
    | jr rs => 
      a =>> (fun P s => exists pi, (codeptr (rget s rs) pi //\\ interp pi) P s) 
  end.

(** well-formed code heap rule  *)
Definition ch_ok (P : chspec) (ch : code_heap) (P' : chspec) : Prop :=
  forall f pi, chspec_in P' f pi ->
    opt_pred (ch f) (fun is => iseq_ok (lift (interp pi) P) is).

(** well-formed program rule  *)
Definition prog_ok (P : chspec) (p : prog) (a : chstpred): Prop :=
  ch_ok P (ch_of p) P /\ a P (st_of p) /\ iseq_ok a (is_of p).

Definition prog_correct (P : chspec) (p : prog): Prop :=
  match (is_of p) with
    | br bt rs rt offset is' => 
      if (btest bt (rget (st_of p) rs) (rget (st_of p) rt))
	then exists pi, (codeptr offset pi //\\ interp pi) P (st_of p)
	else True
    | j target => 
      exists pi, (codeptr target pi //\\ interp pi) P (st_of p)
    | jal fret target => 
      exists pi, (codeptr target pi //\\ interp pi) P (rset (st_of p) r31 fret)
    | jr rs => 
      exists pi, (codeptr (rget (st_of p) rs) pi //\\ interp pi) P (st_of p)
    | _ => True
  end.

(** ******************************************************************* *)
(** soundness proof *)
(** ******************************************************************* *)
Lemma prog_ok_correct : forall P p a,
  prog_ok P p a -> prog_correct P p.
Proof.
  intros P [[ch s] is] a.
  unfold prog_ok; unfold prog_correct; dst.
  intros [Hch [Hs His]].
  destruct is; trivial.

  destruct His as [a' [His' His]]; fold iseq_ok in His'.
  hypreplace His (His _ _ Hs); simpl in * ;  
    destruct (btest b (rget s r) (rget s r0)); trivial.

  hypreplace His (His _ _ Hs); trivial. 
  hypreplace His (His _ _ Hs); trivial. 
  hypreplace His (His _ _ Hs); trivial.
Qed. 

Lemma prog_ok_is_inv : forall P p a,
  prog_ok P p a -> exists p', exists a', step p = Some p' /\ prog_ok P p' a'.
Proof.
  intros P [[ch s] is] a.
  unfold prog_ok; unfold step; unfold ch_of;
    unfold st_of; unfold is_of; simpl.
  intros [Hch [Hs His]].
  destruct is.
  
  destruct His as [a' [Hc His']]; fold iseq_ok in His'.
  hypreplace Hc (Hc _ _ Hs).
  simpl in * .
  destruct (cstep c s) as [s' |]; [idtac | elimtype False; exact Hc].
  exists (ch, s', is); exists a'.
  repeat (split; trivial).
  
  destruct His as [a' [His' His]]; fold iseq_ok in His'.
  hypreplace His (His _ _ Hs); simpl in * ; fold iseq_ok in His. 
  destruct (btest b (rget s r) (rget s r0)).
  assert (Hch' := Hch n).
  destruct His as [pi [Hin Hwf]].
  hypreplace Hch' (Hch' _ Hin).
  simpl in Hch'.
  destruct (ch n) as [is' |]; [idtac | elimtype False; exact Hch'].
  exists (ch, s, is').
  exists (lift (interp pi) P).
  repeat (split; trivial).
  exists (ch, s, is).
  exists a'.
  repeat (split; trivial).
  
  hypreplace His (His _ _ Hs); fold iseq_ok in His. 
  simpl in * .
  assert (Hch' := Hch n).
  destruct His as [pi [Hin Hwf]].
  hypreplace Hch' (Hch' _ Hin).
  simpl in Hch'.
  destruct (ch n) as [is' |]; [idtac | elimtype False; exact Hch'].
  exists (ch, s, is').
  exists (lift (interp pi) P).
  repeat (split; trivial).

  hypreplace His (His _ _ Hs); fold iseq_ok in His. 
  simpl.
  assert (Hch' := Hch n0).
  destruct His as [pi [Hin Hwf]].
  hypreplace Hch' (Hch' _ Hin).
  simpl in Hch'.
  destruct (ch n0) as [is' |]; [idtac | elimtype False; exact Hch'].
  exists (ch, (rset s r31 n), is').
  exists (lift (interp pi) P).
  repeat (split; trivial).
  
  hypreplace His (His _ _ Hs); fold iseq_ok in His. 
  simpl.
  assert (Hch' := Hch (rget s r)).
  destruct His as [pi [Hin Hwf]].
  hypreplace Hch' (Hch' _ Hin).
  simpl in Hch'.
  destruct (ch (rget s r)) as [is' |]; [idtac | elimtype False; exact Hch'].
  exists (ch, s, is').
  exists (lift (interp pi) P).
  repeat (split; trivial).
Qed.
  
Theorem cap_soundness_correct : forall P p a,
  prog_ok P p a -> 
  forall n, exists p', stepn n p = Some p' /\ prog_correct P p'.
Proof.
  intros.
  generalize P p a H; clear P p a H.
  induction n; intros.
  exists p; split; [reflexivity | apply (prog_ok_correct H)].
  simpl.
  destruct (prog_ok_is_inv H) as [p' [a' [Hstep Hok]]].
  rewrite Hstep.
  eapply IHn.
  apply Hok.
Qed.


(** ******************************************************************* *)
(** weakening lemma *)
(** ******************************************************************* *)
Lemma comm_ok_weakening : forall c (p q p' q' : chstpred),
  p =>> p' ->
  q' =>> q ->
  comm_ok p' q' c -> comm_ok p q c.
Proof.
  unfold comm_ok; red; intros.
  hypreplace H1 (H1 _ _ (H _ _ H2)).
  simpl in * .
  destruct (cstep c s);
    [apply (H0 _ _ H1)| elimtype False; exact H1].
Qed.

Lemma iseq_ok_weakening : forall is (a a' : chstpred),
  a =>> a' ->
  iseq_ok a' is -> iseq_ok a is.
Proof.
  induction is; intros.
  destruct H0 as [a'' [Hc His']].
  fold iseq_ok in His'.
  exists a''.
  split; trivial.
  eapply comm_ok_weakening.
  3: apply Hc.
  exact H.
  firstorder.

  destruct H0 as [a0 [His' His]]; fold iseq_ok in His'.
  exists a0.
  split; trivial.
  red; intros. 
  hypreplace His (His _ _ (H _ _ H0)). 
  simpl in * .
  destruct (btest b (rget s r) (rget s r0));
    exact His.
  
  repeat (red; intros).
  hypreplace H0 (H0 _ _ (H _ _ H1)); fold iseq_ok in H0. 
  exact H0.
  
  repeat (red; intros).
  hypreplace H0 (H0 _ _ (H _ _ H1)); fold iseq_ok in H0. 
  exact H0.

  repeat (red; intros).
  hypreplace H0 (H0 _ _ (H _ _ H1)); fold iseq_ok in H0. 
  exact H0.
Qed.


Lemma chspec_subset_iseq_ok : forall is P1 P2 a,
  (forall f T, chspec_in P1 f T -> chspec_in P2 f T) ->
  iseq_ok (lift a P1) is -> 
  iseq_ok (lift a P2) is.
Proof.
  induction is; unfold lift; simpl; intros.
  
  destruct H0 as [a' [Comm Is']].
  exists a'; split; trivial.
  eapply comm_ok_weakening; eauto.
  intros PP ss [In A].
  intuition.
  red; intros; trivial.
  
  destruct H0 as [a' [Is' Is]].
  exists a'; split; trivial.
  intros PP ss [In A].
  apply Is.
  intuition.

  intros PP ss [In A].
  apply H0.
  intuition.

  intros PP ss [In A].
  apply H0.
  intuition.

  intros PP ss [In A].
  apply H0.
  intuition.
Qed.
