(** Lightweighted OCAP. *)

(** Sepcification Constructs. *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export machine2.
Require Export optutil.

(** 
Specialized OCAP for supporting interaction of two verification systems.

An ispec may have one of two kinds, one for TAL, and one for SCAP. 
*)

(** ******************************************************************* *)
(** specification constructs *)
(** ******************************************************************* *)

(** TAL *)
Definition iflg := bool.

Inductive wty : Set :=
  | tvar        : nat -> wty
  | tnull       : wty
  | tint        : wty
  | tcode       : regty -> wty
  | tmem        : iflg -> wty -> iflg -> wty -> wty
  | trec        : wty -> wty
  | tunion      : wty -> wty -> wty
with regty : Set := 
  | regty_nil   : regty
  | regty_cons  : reg -> wty -> regty -> regty
.


(** SCAP *)

(** state predicate *)
Definition spred := state -> Prop.

(** guarantee *)
Definition guar := state -> state -> Prop.

Inductive ispec : Type :=
  | isp1 : regty -> ispec
  | isp2 : spred -> guar -> ispec
.


(** ******************************************************************* *)
(** code heap specification *)
(** ******************************************************************* *)
Definition chspec := listT (prodT nat ispec).

(** in chspec *)
Fixpoint chspec_in (P : chspec) (l : nat) (pi : ispec) {struct P} : Prop :=
  match P with
    | nilT => False
    | consT (pairT l' pi') P' => 
      if (beq_nat l l') then (pi = pi') \/ (chspec_in P' l pi)
	else (chspec_in P' l pi)
  end.


(** get pi from Psi *)
Fixpoint chspec_get1 (P : chspec) (n : nat) {struct P} : optionT regty :=
  match P with
    | consT (pairT n' (isp2 p g)) P' => chspec_get1 P' n
    | consT (pairT n' (isp1 G)) P' => 
      if beq_nat n n' then SomeT G else chspec_get1 P' n
    | nilT => NoneT _
  end.

Fixpoint chspec_get2 (P : chspec) (n : nat) {struct P} : optionT (prodT spred guar) :=
  match P with
    | consT (pairT n' (isp1 G)) P' => chspec_get2 P' n
    | consT (pairT n' (isp2 p g)) P' => 
      if beq_nat n n' then SomeT (pairT p g) else chspec_get2 P' n
    | nilT => NoneT _
  end.

Lemma chspec_get1_dec : forall P n,
  chspec_get1 P n = NoneT _ \/ exists G, chspec_get1 P n = SomeT G.
Proof.
  intros.
  destruct (chspec_get1 P n); intuition.
  right; exists r; reflexivity.
Qed.

Lemma chspec_get2_dec : forall P n,
  chspec_get2 P n = NoneT _ \/ exists bsp, chspec_get2 P n = SomeT bsp.
Proof.
  intros.
  destruct (chspec_get2 P n); intuition.
  right; exists (pairT a b); reflexivity.
Qed.

Lemma chspec_get1_SomeT_in : forall P n G,
  chspec_get1 P n = SomeT G -> chspec_in P n (isp1 G).
Proof.
  induction P; simpl; intros.
  discriminate.
  destruct a as [n' pi].
  destruct pi as [G' | p' g'].
  generalize H; clear H; simplif; subst.
  intro F; injection F; intros; subst; left; reflexivity.
  apply IHP.
  hypreplace H (IHP _ _ H).
  simplif; subst; intuition.
Qed.

Lemma chspec_get2_SomeT_in : forall P n p g,
  chspec_get2 P n = SomeT (pairT p g) -> chspec_in P n (isp2 p g).
Proof.
  induction P; simpl; intros.
  discriminate.
  destruct a as [n' pi].
  destruct pi as [G | p' g'].
  hypreplace H (IHP _ _ _ H).
  simplif; subst; intuition.
  generalize H; clear H; simplif; subst.
  intro F; injection F; intros; subst; left; reflexivity.
  apply IHP.
Qed.

(** get all pis from Psi *)
Fixpoint chspec_get (P : chspec) (l : nat) {struct P} : listT ispec :=
  match P with
    | nilT => nilT _
    | consT (pairT l' pi) P' => 
      if (beq_nat l l') then consT pi (chspec_get P' l)
	else chspec_get P' l
  end.

(** domain of Psi *)
Fixpoint chspec_dom (P : chspec) {struct P} : list nat :=
  match P with
    | nilT => nil
    | consT (pairT l pi) P' => cons l (chspec_dom P')
  end.

Lemma in_chspec_get_chspec_in : forall PP l T,
  InT T (chspec_get PP l) -> chspec_in PP l T.
Proof.
  induction PP as [| [x T'] PP].
  intros; inversion H.

  intros l T; simpl.
  simplif; subst; simpl; intros.
  destruct H; [left | right]; intuition.
  intuition.
Qed.

Lemma dom_get_subset_chspec_subset : forall PP PP',
  (forall l, In l (chspec_dom PP) -> 
    listT_sub (chspec_get PP l) (chspec_get PP' l)) -> 
  (forall f T, chspec_in PP f T -> chspec_in PP' f T).
Proof.
  induction PP as [| [f' T'] PP].
  intros; inversion H0.
  intros PP' Sub f T.
  simpl.
  generalize (Sub f).
  simpl; simplif; subst; intros.
  destruct H0; subst.
  apply in_chspec_get_chspec_in.
  apply H; simpl; try simplbeq; simpl; intuition.
  apply IHPP; trivial.
  intros.
  apply Sub.
  right; trivial.
  simpl.  
  simplif; subst; simpl; intuition.
  apply IHPP; trivial.
  intros.
  apply Sub.
  right; trivial.
  simpl.  
  simplif; subst; simpl; intuition.
Qed.

Lemma chspec_in_in_dom_in_get : forall PP l T,
  chspec_in PP l T -> 
  In l (chspec_dom PP) /\ InT T (chspec_get PP l).
Proof.
  induction PP as [| [x T'] PP].
  intros; inversion H.
  intros l T; simpl.
  simplif; intros.
  destruct H; subst; split; simpl; intuition.
  right; destruct (IHPP _ _ H); trivial.
  destruct (IHPP _ _ H).
  intuition.
Qed.


(** ******************************************************************* *)
(** valid code pointer *)
(** ******************************************************************* *)
Notation codeptr := (fun f pi => (fun P s => chspec_in P f pi)).


(** ******************************************************************* *)
(** assert: code heap and state predicate *)
(** ******************************************************************* *)
Definition chstpred := chspec -> state -> Prop.


(** ******************************************************************* *)
(** assert level logical operators *)
(** ******************************************************************* *)
Definition chstpred_and (A B : chstpred) := (fun P s => A P s /\ B P s).

Infix "//\\" := chstpred_and (at level 61, right associativity).

Definition chstpred_or (A B : chstpred) := (fun P s => A P s \/ B P s).

Infix "\\//" := chstpred_or (at level 62, right associativity).

Definition chstpred_impl (A B : chstpred) := forall P s, A P s -> B P s.

Infix "=>>" := chstpred_impl (at level 63, right associativity).


(** ******************************************************************* *)
(** lift local chspec to globle chspec *)
(** ******************************************************************* *)
Definition lift a P : chstpred :=
  fun P' s => 
    (forall f T, chspec_in P f T -> codeptr f T P' s) /\ (* P \subset P' *)
    (a P' s).