(** Basic TAL type system. *)

(** Typing rules *)

(** Chunxiao Lin @ 2006 *)

Set Implicit Arguments.
Unset Strict Implicit.

Require Export spec.
Require Export reach.


(** types are encoded using de bruijn indices *)

(** any trec is a well-formed type exp, no free variables. *)

(** only need to implement unfold on rec type, that is, subst 0 by t'. *)

Notation rty_cons := (fun R G => regty_cons (fst R) (snd R) G).

Notation "[{ r , .. , r' }]" := 
  (rty_cons r .. (rty_cons r' regty_nil) ..).

Notation "[{}]" := (regty_nil).

Notation "[< T1 ^^ b1 , T2 ^^ b2 >]" := (tmem b1 T1 b2 T2).

Infix "\+/" := tunion (at level 60, right associativity).


(** ******************************************************************* *)
(** rec type functions *)
(** ******************************************************************* *)

(** unfold a rec type *)
Fixpoint rec_unfold_aux (X : nat) (T : wty) (T' : wty) {struct T}: wty :=
  match T with
    | tvar Y => 
      if beq_nat Y X then T' else tvar Y
    | tnull => tnull
    | tint => tint
    | tcode G => 
      tcode (rec_unfold_regty X G T')
    | tmem b1 T1 b2 T2 => 
      tmem b1 (rec_unfold_aux X T1 T') b2 (rec_unfold_aux X T2 T')
    | trec Trec => 
      trec (rec_unfold_aux (S X) Trec T')
    | tunion Ts1 Ts2 => 
      tunion (rec_unfold_aux X Ts1 T') 
      (rec_unfold_aux X Ts2 T')
  end 
with rec_unfold_regty 
(X : nat) (G : regty) (T' : wty) {struct G} : regty :=
match G with
  | regty_nil => regty_nil
  | regty_cons r Tr G' => 
    regty_cons r (rec_unfold_aux X Tr T') (rec_unfold_regty X G' T')
end.
  
Definition rec_unfold (T : wty) : wty :=
  match T with
    | trec T' => rec_unfold_aux 0 T' T
    | _ => T
  end.

(** check all type vars are bound *)
Fixpoint rec_bound_aux (X : nat) (T : wty) {struct T} : bool :=
  match T with
    | tvar Y => (blt_nat Y X)
    | tnull => true
    | tint => true
    | tcode G => rec_bound_regty X G 
    | tmem b1 T1 b2 T2 => 
      andb (rec_bound_aux X T1) (rec_bound_aux X T2)
    | trec Trec => rec_bound_aux (S X) Trec
    | tunion Ts1 Ts2 => andb (rec_bound_aux X Ts1) (rec_bound_aux X Ts2)
  end 
with rec_bound_regty (X : nat) (G : regty) {struct G} : bool :=
match G with
  | regty_nil => true
  | regty_cons r Tr G' => andb (rec_bound_aux X Tr) (rec_bound_regty X G')
end.

Notation rec_bound := (fun T => rec_bound_aux 0 T).

(* Definition list : wty := *)
(* trec (tunion tnull (tmem (cellty_cons true (trec (tunion (tvar 0) (tvar 1))))  *)
(*   (cellty_cons true (tvar 0)))). *)

(* Eval compute in rec_bound list. *)

(* Goal (rec_unfold list) = list. *)
(*   simpl. *)



(** ******************************************************************* *)
(** operations on regty *)
(** ******************************************************************* *)
Fixpoint regty_get (G : regty)  (r : reg) {struct G} : option wty :=
  match G with
    | regty_nil => None
    | regty_cons r' T G' =>
      if (beq_reg r r') then Some T else (regty_get G' r)
  end.

Fixpoint regty_set (G : regty) (r : reg) (T : wty) {struct G} : regty :=
  match G with
    | regty_nil => regty_cons r T regty_nil
    | regty_cons r' T' G' =>
      if (beq_reg r r') then (regty_cons r' T G') else (regty_cons r' T' (regty_set G' r T))
  end.

Fixpoint regty_dom (G : regty) : list reg :=
  match G with
    | regty_nil => nil
    | regty_cons r' T G' => cons r' (regty_dom G')
  end.

Notation regty_indom := (fun r G => rlst_in r (regty_dom G)).

Fixpoint regty_canon (G : regty) : Prop :=
  match G with
    | regty_nil => True
    | regty_cons r T G' => regty_canon G' /\ (~ regty_indom r G')
  end.


(** ******************************************************************* *)
(** TAL used regs, except zero *)
(** ******************************************************************* *)
Notation tal_regs := (s1 :: s2 :: zero :: ra :: nil).

Notation tal_regs' := (s1::s2::ra::nil).

Lemma tal_regs'_intro : forall r,
  rlst_in r tal_regs -> r <> zero -> rlst_in r tal_regs'.
Proof.
  intros.
  firstorder.
  subst; firstorder.
Qed.



(** ******************************************************************* *)
(** heap specification *)
(** ******************************************************************* *)
Definition cellty := prod iflg wty.
(** Phi *)
Definition dhspec := nat -> option (prod cellty cellty).


(** ******************************************************************* *)
(** induction scheme for wty & regty *)
(** ******************************************************************* *)

Scheme wrty_rec :=
  Induction for wty Sort Prop
 with rwty_rec :=
  Induction for regty Sort Prop.


(** ******************************************************************* *)
(** lemmas *)
(** ******************************************************************* *)

(** rec_bound & rec_unfold *)
Lemma rec_bound_fold' : forall T T' n,
  rec_bound_aux n (rec_unfold_aux n T T') = true ->
  rec_bound_aux n (trec T) = true.
Proof.
  intro T; pattern T; apply wrty_rec with
    (P0 := fun G => 
      forall T' n, 
	rec_bound_aux n (rec_unfold_aux n (tcode G) T') = true -> 
	rec_bound_aux n (trec (tcode G)) = true); simpl.
  intros n T' n'.
  compare n n'; intro e; subst; simplbeq.
  intros; apply blt_true; auto.
  simpl; intros.
  hypreplace H (blt_lt H).
  apply blt_true; auto.
  
  trivial.
  trivial.
  trivial.
  
  intros b1 T1 Ind1 b2 T2 Ind2 T' n Ht.
  intros; elimbool.
  splitb; firstorder.

  intros Tr Ind T' n Ht.
  apply Ind with T'; trivial.

  intros T1 Ind1 T2 Ind2 T' n Ht.
  elimbool.
  splitb; firstorder.

  trivial.

  intros r Tw Indw G Indg T' n Hg.
  elimbool.
  splitb; firstorder.
Qed.
  
Lemma rec_bound_fold : forall T,
  rec_bound (rec_unfold (trec T)) = true ->
  rec_bound (trec T) = true.
Proof.
  intros.
  apply rec_bound_fold' with (n := 0) (T := T) (T' := (trec T)).
  apply H.
Qed.

Lemma rec_bound_aux_inc : forall T n, 
  rec_bound_aux n T = true -> rec_bound_aux (S n) T = true.
Proof.
  intro T; pattern T; apply wrty_rec with
    (P0 := fun G => 
      forall n, 
	rec_bound_aux n (tcode G) = true -> 
	rec_bound_aux (S n) (tcode G) = true); simpl.
  intros n n' H.
  hypreplace H (blt_lt H).
  apply blt_true; auto.
  
  trivial.
  trivial.
  trivial.
  
  intros b1 T1 Ind1 b2 T2 Ind2 n Ht.
  intros; elimbool.
  splitb; firstorder.

  intros Tr Ind n Ht.
  apply Ind; trivial.

  intros T1 Ind1 T2 Ind2 n Ht.
  elimbool.
  splitb; firstorder.

  trivial.

  intros r Tw Indw G Indg n Hg.
  elimbool.
  splitb; firstorder.
Qed.
  

Lemma rec_bound_unfold' : forall T T' n,
  rec_bound_aux n (trec T) = true ->
  rec_bound_aux n T' = true ->
  rec_bound_aux n (rec_unfold_aux n T T') = true.
Proof.
  intro T; pattern T; apply wrty_rec with
    (P0 := fun G => 
      forall T' n, 
	rec_bound_aux n (trec (tcode G)) = true ->
	rec_bound_aux n T' = true ->
	rec_bound_aux n (rec_unfold_aux n (tcode G) T') = true); simpl.
  intros n T' n'.
  compare n n'; intro e; subst; simplbeq.
  trivial.
  simpl; intros.
  hypreplace H (blt_lt H).
  apply blt_true; omega.
  
  trivial.
  trivial.
  trivial.
  
  intros b1 T1 Ind1 b2 T2 Ind2 T' n Hb Ht.
  intros; elimbool.
  splitb; firstorder.

  intros Tr Ind T' n Hb Ht.
  apply Ind; trivial.
  apply (rec_bound_aux_inc Ht).

  intros T1 Ind1 T2 Ind2 T' n Hb Ht.
  elimbool.
  splitb; firstorder.

  trivial.

  intros r Tw Indw G Indg T' n Hb Hg.
  elimbool.
  splitb; firstorder.
Qed.

Lemma rec_bound_unfold : forall T,
  rec_bound (trec T) = true ->
  rec_bound (rec_unfold (trec T)) = true.
Proof.
  intros.
  apply rec_bound_unfold' 
    with (n := 0) (T := T) (T' := (trec T));apply H.
Qed.

(** regty *)
Lemma regty_indom_intro : forall G r w, 
  regty_get G r = Some w -> regty_indom r G.
Proof.
  induction G; simpl; intros.
  discriminate.
  simpl.
  comparer r0 r; [left | right].
  reflexivity.
  rewrite (beq_reg_intro2 e) in H.
  firstorder. 
Qed.

Lemma regty_indom_elim : forall G r, 
  regty_indom r G -> exists w, regty_get G r = Some w.
Proof.
  induction G; simpl; intros.
  tauto.
  destruct H; subst.
  solvereg.
  exists w; reflexivity.
  destruct (IHG _ H).
  rewrite H0.
  destruct (beq_reg r0 r); [exists w | exists x]; reflexivity.
Qed.

Lemma regty_set_get_eq : forall G r T,
  regty_get (regty_set G r T) r = Some T.
Proof.
  induction G; simpl; intros.
  solvereg.
  reflexivity.
  comparer r0 r; simpl; solvereg; 
    simpl; solvereg; trivial.
Qed.

Lemma regty_set_get_neq : forall G r r' T,
  r' <> r -> regty_get (regty_set G r T) r' = regty_get G r'.
Proof.
  induction G as [ | rt]; intros.
  simpl; solvereg.
  reflexivity.
  simpl.
  comparer r rt; solvereg.
  simpl; solvereg.
  reflexivity.
  simpl; solvereg.
  destruct (beq_reg r' rt); auto.
Qed.

Lemma rec_bound_regty_intro : forall G,
  regty_canon G ->
  (forall r T, regty_get G r = Some T -> rec_bound T = true) -> 
  rec_bound_regty 0 G = true.
Proof.
  induction G; intros.
  reflexivity.
  simpl.
  splitb.
  apply (H0 r w).
  simpl. 
  solvereg.
  reflexivity.
  apply IHG; intros; simpl in H.
  firstorder.
  comparer r0 r; solvereg.
  hypreplace H1 (regty_indom_intro H1).
  firstorder.
  apply (H0 r0 T).
  simpl. 
  solvereg.
  apply H1.
Qed.

Lemma rec_bound_regty_elim : forall G,
  regty_canon G ->
  rec_bound_regty 0 G = true ->
  (forall r T, regty_get G r = Some T -> rec_bound T = true).
Proof.
  induction G.
  intros.
  simpl in H1.
  discriminate.
  simpl; intros Hc Hb rt T.
  comparer rt r; solvereg; intros; elimbool.
  injection H; intros; subst.
  apply B.
  apply IHG with rt;
    firstorder.
Qed.

Lemma regty_set_dom_subset : forall G r T,
  rlst_subset (regty_dom G) (regty_dom (regty_set G r T)).
Proof.
  induction G; simpl; red; intros.
  firstorder.
  comparer r0 r; solvereg;
  simpl in H; destruct H; subst;firstorder.
Qed.

Lemma regty_set_dom_subset2 : forall G r rt T,
  r <> rt -> rlst_in r (regty_dom (regty_set G rt T)) -> rlst_in r (regty_dom G) .
Proof.
  induction G; simpl.
  intros; firstorder.
  intros r0 rt T F.
  comparer rt r; solvereg; simpl.
  firstorder.
  intro H; destruct H.
  firstorder.
  right.
  apply (IHG _ _ _ F H).
Qed.

Lemma regty_canon_rset : forall G r T,
  regty_canon G -> regty_canon (regty_set G r T).
Proof.
  induction G; simpl; intros.
  firstorder.
  destruct H.
  comparer r0 r; solvereg.
  firstorder.
  split.
  firstorder.
  intro F.
  apply H0.
  apply regty_set_dom_subset2 with r0 T; auto.
Qed.


(** ******************************************************************* *)
(** well-formed type *)
(** ******************************************************************* *)

Definition WFwty (T : wty) := rec_bound T = true.

Notation WFzero := (fun G =>
  (regty_get G zero = Some tnull \/ regty_get G zero = Some tint \/ regty_get G zero = None)).

Definition WFregty (G : regty) := 
  rec_bound_regty 0 G = true /\ (* no free var *)
  rlst_subset (regty_dom G) tal_regs /\ (* domain is bound *)
  regty_canon G /\ (* no duplication *)
  WFzero G (* well-formed zero reg *)
. 

Definition WFdhspec (Psi : dhspec) :=
  forall l b1 T1 b2 T2, 
    Psi l = Some ((b1, T1), (b2, T2)) -> 
    vptr l /\ WFwty T1 /\ WFwty T2.

(** ******************************************************************* *)
(** subtyping  *)
(** ******************************************************************* *)

(** heap cell subtyping *)
Inductive cellty_sub : cellty -> cellty -> Prop :=
  | cellty_sub_refl : forall C, cellty_sub C C
  | cellty_sub_tf : forall T, 
    cellty_sub (true, T) (false, T)
.

(** word value subtyping *)
Inductive wty_sub : wty -> wty -> Prop :=
  | wty_sub_null : wty_sub tnull tint
  | wty_sub_ul : forall T T', WFwty T' -> wty_sub T (tunion T T')
  | wty_sub_ur : forall T T', WFwty T -> wty_sub T' (tunion T T')
.

(** register subtyping *)
Inductive regty_sub : regty -> regty -> Prop :=
  | regty_sub_dom : forall G G',
    (forall r, regty_indom r G' -> regty_get G' r = regty_get G r) ->
    WFregty G' -> regty_sub G G'
  | regty_sub_fold : forall G r T,
    r <> zero ->
    regty_get G r = Some (rec_unfold (trec T)) ->
    WFregty G ->
    regty_sub G (regty_set G r (trec T))
  | regty_sub_unfold : forall G r T,
    r <> zero ->
    regty_get G r = Some (trec T) -> 
    WFregty G ->
    regty_sub G (regty_set G r (rec_unfold (trec T)))
.

(** ******************************************************************* *)
(** word value typing *)
(** ******************************************************************* *)
Inductive WFw : chspec -> dhspec -> nat -> wty -> Prop :=
  | WFw_null : forall Phi Psi, WFw Phi Psi 0 tnull
  | WFw_int  : forall Phi Psi i, WFw Phi Psi i tint

  | WFw_code : (* code label *)
    forall Phi Psi l G, 
      chspec_in Phi l (isp1 G) -> 
      WFw Phi Psi l (tcode G)

  | WFw_mem  : (* memory pointer *)
    forall Phi Psi addr C1 C2 b1 T1 b2 T2, 
      Psi addr = Some (C1, C2) -> 
      cellty_sub C1 (b1,T1) ->
      cellty_sub C2 (b2,T2) ->
      WFw Phi Psi addr (tmem b1 T1 b2 T2)

  | WFw_fold : (* fold rec type *)
    forall Phi Psi w T, 
      WFw Phi Psi w (rec_unfold (trec T)) ->
      WFw Phi Psi w (trec T)

  | WFw_sub : (* subtyping *)
    forall Phi Psi w T T', 
      WFw Phi Psi w T -> wty_sub T T' -> WFw Phi Psi w T'
.

(** heap value typing *)
Inductive WFcell : chspec -> dhspec -> nat -> cellty -> Prop :=
  | WFcell_junk : 
    forall Phi Psi w T,
      WFcell Phi Psi w (false, T)
  | WFcell_init : 
    forall Phi Psi w T,
      WFw Phi Psi w T ->
      WFcell Phi Psi w (true, T)
.

(** ******************************************************************* *)
(** reg file typing *)
(** ******************************************************************* *)
Definition WFrf (Phi : chspec) (Psi : dhspec) (rf : rfile) (G : regty) : Prop :=
  (forall r T, regty_get G r = Some T -> WFw Phi Psi (rfile_get rf r) T) /\ WFregty G.


(** ******************************************************************* *)
(** heap typing *)
(** ******************************************************************* *)

Definition WFdh (Phi : chspec) (Psi : dhspec) (hp : heap) (Psi' : dhspec) : Prop :=
  (forall l C1 C2, Psi' l = Some (C1, C2) -> 
    exists v1, exists v2,
      heap_get hp l = Some v1 /\ 
      heap_get hp (l + WSIZE) = Some v2 /\ 
      WFcell Phi Psi v1 C1 /\
      WFcell Phi Psi v2 C2) /\ 
  WFdhspec Psi'.


(** ******************************************************************* *)
(** state typing *)
(** ******************************************************************* *)
Definition WFst (Phi : chspec) (G : regty) (st : state) : Prop :=
  let hp := hp_of st in
    let rf := rf_of st in
      exists Psi : dhspec,
	WFdh Phi Psi hp Psi /\ WFrf Phi Psi rf G.
 


(* (\** ******************************************************************* *\) *)
(* (\** well-formed code heap *\) *)
(* (\** ******************************************************************* *\) *)

(* Definition WFcs (Psi : chspec) : Prop := *)
(*   forall f G, chspec_in Psi f (isp1 G) -> WFregty G. *)
  
(* Definition WFch (Psi : chspec) (ch : code_heap) (Psi' : chspec) : Prop := *)
(*   forall f G, Psi' f = Some G -> *)
(*     opt_pred (ch f) (fun is => WFis Psi G is). *)


(* (\** ******************************************************************* *\) *)
(* (\** well-formed program *\) *)
(* (\** ******************************************************************* *\) *)
(* Definition WFprog (Psi : chspec) (G : regty) (p : prog) : Prop := *)
(*   let ch := ch_of p in *)
(*     let is := is_of p in *)
(*       let st := st_of p in *)
(* 	WFcs Psi /\ WFch Psi ch Psi /\ WFst Psi G st /\ WFis Psi G is. *)